include("AI_init.h").
include("AI_window.h").
include("AI_data.h").

?-
  G_M := 27, % Models
  G_S := 20, % States (17)
  G_E := 70, % Events
  G_Tr := 14, % Traces
  G_T := 18, % Events+Traces
  G_Sel := 0,
  G_Sel_Tr := -1,
  G_Sel_Tr_H := 1,
  G_Sel_Show := 0,
  G_Pause := 0,
  G_Freeze := 0,
  G_with_Randomness := 0,
  G_Observed := 0,
  G_ObserveNow := 0,
  G_One_Step := 0,
  G_FastSpeed := 0,
  G_Last_Speed := 0,
  G_Results := 0,
  G_Max1 := min_integer,
  G_Max2 := min_integer,
  array(enter, G_M, -1),
  array(exit, G_M, -1),
  array(active, G_M, 1),
  array(restricted, G_M, 0),
  array(time_last, G_M, 0),
  array(used_states, G_M, 0),
  array(all_arrows, G_M, 0),
  array(all_traces, G_M, 0),
  array(active_previous, G_M, 0),
  array(active_last, G_M*2, 1),
  array(events_last, G_M*2, 0),
  array(links, G_M*G_S*G_S, 0),
  array(states, G_M*G_S*G_Tr, 0),
  array(events, G_M*G_T, -1),
  array(events2, G_E*10, max_integer),
  array(events_val, G_E, 0),
  array(results, G_M*8*4, 0), % G_K =< 4
  array(results0, 8*4, 0), % G_K =< 4
  array(buf_views, 50, 0),
  array(buf_incorrect, 50, 0),
  array(buf_actions, 50, 0),
  array(buf_arrows, G_M*50, 0),
  array(buf_the_trace, G_M*50, 0),
  array(board, 66, 0),
  init_all,
  size(screen, _, H),
  (H >= 1080 ->
    G_Zoom_Font := 1,
    G_ChS := 15
  else
    G_Zoom_Font := 0.8,
    G_ChS := 12
  ),
  font(G_Zoom_Font*10, G_Zoom_Font*25, "Courier"),
  (G_Zoom_Font=:=1 -> X := 1270, Y := 1100 else X := 1020, Y := 760),
  window(title("AI Unravels the Chess"), size(X, Y), pos(100, 10), paint_indirectly).

init_all:-
  start_random_from(3), % make the random deterministic
  G_K := 3, % possible values in input/output, from 2 to 4
  G_Time := 0,
  G_Incorrect := 0,
  clean_buffers,
  init_board,
  update_memory,
  board(65):= 7,
  fixed_events,
  fixed_models,
  clean_results,
      %active(1) := 8, active_previous(1) := 8,
      %active(2) := 32, active_previous(2) := 32,
      %board(8*5 + 4):= 5,
      %board(8*5 + 2):= 4,

  for( Model, 16, 25),
    init_model(Model),
    fail.
init_all.

time_func(end) :-
  check_pause,
  enter_or_exit_algorithm(exit, 0),
  update_memory,
  clean_results,
%                 zero (fourth) restriction
  make_restriction(4),
  G_Time := G_Time + 1,
  set_window_text,
%                 observation given by the world
  G_Max1 := min_integer,
  G_Max2 := min_integer,
  evaluete_events(_, forcast),
  calc_observation(0),
  Min1 := min_integer,
  create_probability(Min1, _, _, _, _, Value, calc),
  buf_views(G_Time mod 50):= Value,
  make_restriction(1),
%                 first transition
  evaluete_events(_, transition),
  transition(1),
  make_restriction(2),
%                 action done by the agent
  G_Incorrect := 0,
  evaluete_events(_, forcast),
  calc_observation(1),
  create_incorrect,
  (G_Incorrect =:= 2**G_K - 1 -> find_possible),
  buf_incorrect(G_Time mod 50):= G_Incorrect,
  (G_Freeze =:= 0 ->
    find_correct(Action)
  else
    Action := 0
  ),
  buf_actions(G_Time mod 50):= Action,
  make_restriction(3),
%                 second transition
  evaluete_events(_, transition),
  transition(-1),

  enter_or_exit_algorithm(enter, 1),
  update_window(_).

enter_or_exit_algorithm(What, N):-
  for(Model, 0, G_M-1),
  Event2:= What(Model),
  Event2 =\= -1,
  Cell := events2(10*Event2),
  Delay := true_value((Cell//(2*G_S)) mod 10 =:= 2),
  (event_is_true(Event2, 0, Delay)->
    active_previous(Model):=N,
    active(Model):=N
  ),
  fail.
enter_or_exit_algorithm(What, N).

transition(Views):-
  for( Model, 0, G_M-1),
    transition2(Model, Views),
    fail.
transition(Views).

transition2(Model, Views):-
  events_are_true(Model, Events, Views),
  choose_new_states(Model, Moved, New, Events, New_E),
  (New =\= 0 ->
    New2 := New \/ active(Model) /\ \Moved, 
    (Views =:= 1->
      active_last(2*Model+0) := active(Model),
      active_last(2*Model+1) := New2,
      events_last(2*Model+0) := New_E,
      events_last(2*Model+1) := 0
    else
      (G_Time =\= time_last(Model) ->
        active_last(2*Model+0) := active(Model),
        active_last(2*Model+1) := active(Model),
        events_last(2*Model+0) := 0
      ),
      events_last(2*Model+1) := New_E
    ),
    buf_arrows(50*Model + G_Time mod 50):= events_last(2*Model+0) \/ events_last(2*Model+1),
    time_last(Model) := G_Time,
    active(Model) := New2
  else (Views =:= 1->
    buf_arrows(50*Model + G_Time mod 50):= 0
  )).

choose_new_states(Model, Moved, New, Events, New_E):-
  Moved := 0,  New := 0, New_E := 0,
  Sn := find_any_1_in(active(Model)),
    for(I, 0, G_S-1),
      E := links(G_S*G_S*Model+G_S*Sn+I)/\Events,
      E =\= 0,
        Moved := Moved\/2**Sn,
        New := New\/2**I,
        New_E := New_E\/E,
        fail.
choose_new_states(Model, Moved, New, Events, New_E).

create_incorrect:-
  for( Action, 0, G_K-1),
    (results0(4*(Action+4)) < G_Max2 ->
      G_Incorrect := G_Incorrect \/ 2**Action
    else (results0(4*(Action+4)) =:= G_Max2, results0(4*(Action+4) + 2) < 0 ->
      (random(20) < 5 -> G_Incorrect := G_Incorrect \/ 2**Action)
    )),
    fail.
create_incorrect.

find_possible:-
  for(Action, 0, G_K - 1),
   results0(4*(Action+4)) =:= G_Max2,
   G_Incorrect := 2**G_K - 1 - 2**Action.

find_correct(Action):-
  Action := random(G_K),
  (G_Incorrect /\ 2**Action =\= 0 ->
    find_correct(Action)
  ).

create_probability(Min1, Min2, Number, NumberMin2, Row, Text, _):-
  Number := 0,
  Min2 := max_integer,
  for(I, 0, G_K-1),
    results0(4*I) =:= G_Max1,
    results0(4*I+2) > Min1,
    Number := Number + 1,
    (Min2 > results0(4*I + 2) ->
      NumberMin2 := 1,
      Min2 := results0(4*I + 2)
    else (Min2 =:= results0(4*I + 2) ->
      NumberMin2 := NumberMin2 + 1
    )),
    fail.
create_probability(_, Min2, Number, NumberMin2, Row, Text, paint):-
  (Min2 < results0(4*Row + 2) ->
    Text := Text + "("+(2*Number-NumberMin2)+"/"+2*Number+")*",
    create_probability(Min2, _, _, _, Row, Text, paint)
  else (Number =\= NumberMin2 ->
    Text := Text + "(1/"+2*Number+")"
  else (Number =\= 1 ->
    Text := Text + "(1/"+Number+")"
  else
    Text := Text + "one"
  ))).
create_probability(_, Min2, Number, NumberMin2, _, Value, calc):-
  (Number =:= NumberMin2 ->
    ItsNumber := (G_with_Randomness =:= 0 -> 0 else random(Number)),
    find_its_number(Min2, ItsNumber, Value)
  else
    ItsNumber := (G_with_Randomness =:= 0 -> 2*Number-1 else random(2*Number)),
    (ItsNumber < NumberMin2 ->
      find_its_number(Min2, ItsNumber, Value)
    else
      create_probability(Min2, _, _, _, _, Value, calc)
    )
  ).

find_its_number(Min1, ItsNumber, Value):-
  Number := 0,
  for(I, 0, G_K-1),
    results0(4*I) =:= G_Max1,
    results0(4*I+2) =:= Min1,
    Number := Number + 1,
    ItsNumber < Number,
    Value := I.

make_restriction(Views):-
  evaluete_events(Views, restriction),
  for( Model, 0, G_M-1),
    State := find_any_1_in(active(Model)),
      for(Event, 0, all_traces(Model)-1),
        Trace := states(G_Tr*G_S*Model+G_Tr*State + Event),
        Trace < 3,
        Event2 := events(G_T*Model+Event+4),
        events_val(Event2) > 0,
        (Trace + events_val(Event2)) mod 2=:=1,
        New := active(Model)/\ \2**State,
        (New=\=0->
          restricted(Model):=restricted(Model)\/2**State,
          active(Model):=New
        ),
        fail.
make_restriction(_).

calc_observation(Action):-
  for( Model, 0, G_M-1),
    observable(Model),
    calc_observation2(Model, Action).
calc_observation(Action):-
  for(Value, 4*Action, 4*Action+G_K-1),
    for(I, 0, 1),
      New := 0,
      New1 := 0,
      add_results(New, New1, Value, I),
      results0(4*Value+2*I) := New,
      results0(4*Value+2*I+1) := New1,
      (Action =:= 0 -> 
        (G_Max1 < New, I =:= 0 -> G_Max1 := New)
      else
        (G_Max2 < New, I =:= 0 ->
          G_Max2 := (New < 0 -> New else 0)
        )
      ),
      fail.
calc_observation(Action).

calc_observation2(Model, Action):-
  for(Event, 0, all_traces(Model)-1),
    Trace := states(G_Tr*G_S*Model+G_Tr*find_first_1_in(active(Model)) + Event),
    if_all_agree(Model, active(Model), Event, Trace),
    Trace < 5,
    Event2 := events(G_T*Model+Event+4),
    ((events_val(Event2) mod 32)//8=:=Action+1->
      Value := (events_val(Event2) mod 8)//2,
      Add := 32*Model+16*Action+4*Value+Trace-1,
      ((events_val(Event2)/\32)=:=0 ->
        (events_val(Event2) mod 2=:=1 ->
          Add := (Trace mod 2=:=1 -> Add+1 else Add-1)
        ),
        results(Add) := results(Add) + 1
      else
        (Trace mod 2=:=1 ->
          Add := Add+1
        ),
        ((events_val(Event2)+Trace) mod 2=:=0 ->
          results(Add) := results(Add) + 1
        else
          add_to_all(Add-4*Value),
          results(Add) := results(Add) - 1
        )
      ),
      buf_the_trace(50*Model + G_Time mod 50) := buf_the_trace(50*Model + G_Time mod 50) \/ 2**Event
    ),
    fail.

if_all_agree(Model, States, Event, Trace):-
  State := find_any_1_in(States),
  Trace2 := states(G_Tr*G_S*Model+G_Tr*State+Event),
  Trace =\= Trace2,
  (Trace2<3->
    Trace =\= Trace2+2
  else (Trace2<5->
    (Trace =:= Trace2-2 ->
      Trace := Trace2,
      fail
    )
  )),
  !,
  Trace := 5.
if_all_agree(Model, States, Event, Trace).

add_to_all(Add):-
  for(Value, 0, G_K-1),
    results(Add+4*Value) := results(Add+4*Value) + 1,
    fail.
add_to_all(Add).

add_results(New, New1, Value, I):-
  for( Model, 0, G_M-1),
    New := New + results(32*Model+4*Value+2*I) - results(32*Model+4*Value+2*I+1),
    New1 := New1 + results(32*Model+4*Value+2*I+1),
    fail.
add_results(New, New1, Value, I).

clean_buffers:- for(I, 0, G_M-1), active(I):=true_value(I<5; 15<I), active_previous(I):=1, time_last(I):=0, used_states(I):=1, fail.
clean_buffers:- for(I, 0, G_M*2-1), active_last(I):=1, events_last(I):=0, fail.
clean_buffers:- for(I, 0, G_M*G_S*G_S-1), links(I):=0, fail.
clean_buffers:- for(I, 0, G_M*G_S*G_Tr-1), states(I):=0, fail.
clean_buffers:- for(I, 0, 50-1), buf_views(I):=0, buf_incorrect(I):=0, buf_actions(I):=0, fail.
clean_buffers:- for(I, 0, G_M*50-1), buf_arrows(I):=0, buf_the_trace(I):=0, fail.
clean_buffers:- for(I, 0, G_M*8-1), events(I):= -1, fail.
clean_buffers:- for(I, 0, G_E*10-1), events2(I):=max_integer, fail.
clean_buffers:- for(I, 0, 66-1), board(I):=0, fail.
clean_buffers:- active_previous(G_M-1):=0.

clean_results:- for(I, 0, G_M*8*4-1), results(I):=0, fail.
clean_results:- for(I, 0, 8*4-1), results0(I):=0, fail.
clean_results:-
  for(Model, 0, G_M-2),
    active_previous(Model) := active(Model),
    restricted(Model) := 0,
    buf_the_trace(50*Model + G_Time mod 50) := 0,
    fail.
clean_results.

init_model(Model):-
  N := random(2),
  E := random(3), 
  all_arrows(Model) := 3+N,
  all_traces(Model) := 4, % for trace,  from 1 to 4
  used_states(Model) := 4+random(5),
  (N =:= 0 ->
    events(G_T*Model+0):= 1+E,
    events(G_T*Model+1):= 1+E+3,
    events(G_T*Model+2):= 0
  else
    events(G_T*Model+0):= 1+E,
    events(G_T*Model+1):= 4+E,
    events(G_T*Model+2):= 1 + (E+1) mod 3,
    events(G_T*Model+3):= 4 + (E+1) mod 3
  ),
  K1 := random(3),
  K2 := random(3),
  events(G_T*Model+4):= 35+K1,
  events(G_T*Model+5):= 35+(K1+1) mod 3,
  events(G_T*Model+6):= 38+K2,
  events(G_T*Model+7):= 38+(K2+1) mod 3,
  %init_events(Model, 0, 4, all_traces(Model)),
  init_Links(Model).

init_Links(Model):-
  Used := used_states(Model),%nl,
  add_one_more(Model, Used),
  add_one_more(Model, Used),
  add_one_more(Model, Used),
  for(I, 0, Used-1),
    make_trace(Model, I, 0),
    make_trace(Model, I, 1),
    make_trace(Model, I, 2),
    make_trace(Model, I, 3),
    J := (I+1) mod Used,
    Event := random(4),
    (all_arrows(Model)=:=3, Event>1 ->
      Event := 2
    else
      E2 := (2/\Event)+ (Event+1) mod 2,
      J2 := (I+1+random(Used-1)) mod Used,
      links(G_S*G_S*Model+G_S*I+J2) := links(G_S*G_S*Model+G_S*I+J2)\/2**E2
    ),
    links(G_S*G_S*Model+G_S*I+J) := links(G_S*G_S*Model+G_S*I+J)\/2**Event,
    fail.
init_Links(Model).

add_one_more(Model, Used):-
    I := random(Used),
    J := (I+2+random(2)) mod Used,
    Event := random(2),
    links(G_S*G_S*Model+G_S*I+J) := links(G_S*G_S*Model+G_S*I+J)\/2**Event.
  
make_trace(Model, State, Event):-
  Trace := random(10)+1,
  (Trace>5 -> Trace :=5),
  states(G_Tr*G_S*Model+G_Tr*State+Event) := Trace.

op( 700, xfx, ':==').

[] :== []/S. 
[H|T] :== [H1|T1]/S :- 
  H:= H1/S,
  T:== T1/S.

[A2, B2] :== roll(Alfa, [A, B]) :- 
  A2 := A*cos(Alfa)+B*sin(Alfa),
  B2 := B*cos(Alfa)-A*sin(Alfa).

[] :== []+S*[]. 
[H|T] :== [H1|T1]+S*[H2|T2] :- 
  H:=H1+S*H2,
  T:==T1+S*T2.

