fixed_events:-
  (conjunction(N, _, _, P_Event); implication(N, _, _, P_Event)),
  init_one_fixed_event(N, P_Event, 0),
  fail.
fixed_events.

calc_type(a, 0).
calc_type(cor, 1).
calc_type(v, 2).
calc_type(m, 3).
calc_type(ob, 4).
calc_type(in, 5).
calc_type(event, 6).

init_one_fixed_event(N, event(p( Atom, Type, Value, Negation)|Tail), Count):-
  calc_type(Type, Type2),
  events2(10*N+Count) := 10*G_S*2*Atom +G_S*2*Type2 + 2*Value + true_value(Negation=no),
  init_one_fixed_event(N, event(|Tail), Count+1).
init_one_fixed_event(N, event(), Count):-
  events2(10*N+Count) := max_integer.

fixed_models:-
  fixed_models_data(Model, states(AllStates), Algorithm, Arrows, Traces, Links, StatesTrace),
  all_arrows(Model) := list_length(Arrows), % for arrows, from 1 to 4
  all_traces(Model) := list_length(Traces), % for trace,  from 0 to G_Tr=10
  init_algorithm(Model, Algorithm),
  init_fixed_events(conjunction, Model, 0, 0, Arrows),
  init_fixed_events(implication, Model, 0, 4, Traces),
  init_fixed_Links(Model, Links),
  init_fixed_States(Model, AllStates),
  init_fixed_Trace(Model, StatesTrace),
  fail.
fixed_models.

init_algorithm(Model, Algorithm):-
  (Algorithm=[Enter, Exit]->
    conjunction(Event1, Enter, _, _),
    enter(Model):=Event1,
    conjunction(Event2, Exit, _, _),
    exit(Model):=Event2
  ).

init_fixed_States(Model, AllStates):-
  for(I, 0, AllStates-1),
    for(J, 0, G_Tr-1),
      states(G_Tr*G_S*Model+G_Tr*I+J) := 5,
      fail. 
init_fixed_States(Model, AllStates).

init_fixed_events(Type, Model, Event, Start, [Ind | List] ):-
  Type(N, Ind, Name, P_Event),
  events(G_T*Model+Event+Start):= N,
  init_fixed_events(Type, Model, Event+1, Start, List).
init_fixed_events(Type, Model, Event, Start, []).

init_fixed_Trace(Model, [trace(State, Event, Trace) | List] ):-
  states(G_Tr*G_S*Model+G_Tr*State+Event):= Trace,
  init_fixed_Trace(Model, List).
init_fixed_Trace(Model, []).

init_fixed_Links(Model, [link(State, State2, Event) | List] ):-
  links(G_S*G_S*Model+G_S*State+State2) := links(G_S*G_S*Model+G_S*State+State2)\/2**Event,
  init_fixed_Links(Model, List).
init_fixed_Links(Model, []).

%  fixed_models_data(Model, states(AllStates), [Enter, Exit], Arrows, Traces, Links, StatesTrace),
% event(p( Atom, Type, Value, Negation)|Tail)
% link(State, State2, Event)
% trace(State, Event, TraceVol)

init_board:-
  for(I, 0, 7),
    init_squre(0, I, I+2),
    init_squre(1, I, 1+14),
    fail.
init_board:-
  for(I, 0, 2),
    init_squre(0, 7-I, I+2),
    fail.
init_board:-
  init_squre(0, 0, 2+14),
  init_squre(0, 4, 6+14),
  init_squre(0, 7, 2+14).

init_squre(Line, Column, Piece):-
    board(8*Line + Column):= Piece,
    board(56-8*Line + Column):= Piece+7.

make_chess_squre(X, I, J):-
  (board(8*J+I) =:= 7 ->
    color_text_back(_, rgb(255,255,0))
  else ((I+J) mod 2 =:=0 ->
    color_text_back(_, rgb(0,255,255))
  else
    color_text_back(_, rgb(240,240,240))
  )),
  (board(8*J+I) mod 14 > 6 ->
    color_text(_, rgb(0,0,0))
  else
    color_text(_, rgb(255,0,0))
  ),
  Piece := board(8*J+I) mod 7,
  (J =:= 9, active_previous(3) =:= 1 -> Piece := 0),
  face(Piece, Face),
  (I=:=0 -> 
    text_out(Face, pos(G_Zoom_Font*X, G_Zoom*(925-25*J)))
  else
    text_out(Face)
  ).

show_grid(I, J):-
  pen(3, rgb(0, 0, 0)),
  font(G_Zoom_Font*8, G_Zoom_Font*22, "Courier"), 
  X := G_Zoom*(300 - 35*I),
  Y := G_Zoom*(400 - 35*J),
  R := 70*G_Zoom,
  Step := 20*G_Zoom,
  color_text(_, rgb(0,0,255)),
  horizontal_grid(X, Y, I, J, R),
  vertical_grid(X, Y, I, J, R),
  color_text(_, rgb(0,0,0)),
  for(B, 0, J-1),
    for(A, 0, I-1),
      (G_Sel =:= 5 ->
        Object := board(8*(7-B)+A)
      else
        Object := board(64+A)
      ),
      Object > 0,
      (Object =:= 7 ->
        color_text(_, rgb(0,255,0)),
        names(24, _, _, _, Name),
        text_out(Name, pos(2+X+A*R, 3+Y+B*R)),
        fail
      else (Object mod 14 > 6 ->
        names(17, _, _, _, Name),
        color_text(_, rgb(0,0,0))
      else
        names(16, _, _, _, Name),
        color_text(_, rgb(255,0,0))
      )),
      text_out(Name, pos(2+X+A*R, 3+Y+B*R)),
      color_text(_, rgb(0,0,255)),
      Piece := 17+ Object mod 7,
      names(Piece, _, _, _, Name2),
      text_out(Name2, pos(2+X+A*R, 3+Step+Y+B*R)),
      Object > 14,
      color_text(_, rgb(0,0,0)),
      names(25, _, _, _, Name3),
      text_out(Name3, pos(2+X+A*R, 3+2*Step+Y+B*R)),
      fail.
show_grid(I, J):-
  font(G_Zoom_Font*10, G_Zoom_Font*25, "Courier"). 

horizontal_grid(X, Y, I, J, R):-
  for(B, 0, J),
    line(X, Y+B*R, X+I*R, Y+B*R),
    B < J, J > 1,
    text_out(J-B-1, pos(X-G_Zoom*17, G_Zoom*25+Y+B*R)),
    fail.
horizontal_grid(X, Y, I, J, R).

vertical_grid(X, Y, I, J, R):-
  for(A, 0, I),
    line(X+A*R, Y, X+A*R, Y+J*R),
    A < I, I > 1,
    text_out(A, pos(X+G_Zoom*30+A*R, Y+G_Zoom*5+J*R)),
    fail.
vertical_grid(X, Y, I, J, R).

observable(Atom) :-
  (Atom > 6, Atom < 26 ->
    (G_Observed > 0 ->
      Atom =:= G_Observed
    else (Atom > 15 ->
      Squre := G_ObserveNow,
      Squre > 0,
      (Atom < 18 -> % colour
        (Squre mod 14 > 6 -> Atom=:=17, Squre=\=7 else Atom=:=16)
      else (Atom < 24 -> % piece
        Atom - 17 =:= Squre mod 7
      else
        (Atom=:=24, Squre =:=7; Atom=:=25, Squre > 14)
      ))
    else
      active(Atom)=\=0
    ))
  ).

up:- % m0=2 & a0=a
  active_previous(0)=:=4, % 2**2
  buf_actions(G_Time mod 50)=:=1.

down :- % m0=2 & a0=b
  active_previous(0)=:=4,
  buf_actions(G_Time mod 50)=:=2.

here(I, J):-
  I := find_first_1_in(active_previous(1)),
  J := find_first_1_in(active_previous(2)).

somewhere(I, J):-
  repeat,
  I := random(8),
  J := random(8).

update_memory :- % up, here => copy(Lifted)
  up,
  here(I, J),
  board(8*J+I) := 7,
  fail.
update_memory :- % down, everywhere => del(Lifted)
  down,
  for(I, 0, 7),
    for(J, 0, 7),
      (board(8*J+I) =:= 7 -> board(8*J+I) := 0),
      fail.
update_memory :- % down, here => copy(mem5(0))
  down,
  here(I, J),
  board(8*J+I) := board(64+0),
  fail.
update_memory :- % down, !ob(Lifted), here => del(Unmoved)
  down,
  G_ObserveNow =\= 7,
  here(I, J),
  board(8*J+I) := board(8*J+I) mod 14,
  fail.
update_memory :- %  => ob(now) := mem4(here)
  I := find_first_1_in(active(1)),
  J := find_first_1_in(active(2)),
  G_ObserveNow := board(8*J+I),
  fail.
/*----------------*/
update_memory :- %  => mem5(here) := ob(now)
  I := find_first_1_in(active(3)),
  board(64+I) := G_ObserveNow,
  fail.
update_memory :- % down, where=1 => copy(Lifted)
  down,
  I := 1,
  board(64+I) := 7,
  fail.
update_memory.

show_memory_rules:-
  I := 2,
  memory_rules(Number, G_Sel, Text1, Text2),
  I := I + 1,
  color_text(_, rgb(255,0,0)),
  text_out(Number, pos(G_Zoom*600, G_Zoom*(30+I*40))),
  color_text(_, rgb(0,0,0)),
  text_out(Text1),
  color_text(_, rgb(0,0,255)),
  text_out(Text2).

evaluete_events(_, _):- % transition and ...
  conjunction(Event, Ind, Name, P_Event),
  Cell := events2(10*Event),
  (Cell =:= max_integer ->
    events_val(Event) := 5
  else (event_is_true(Event, 0, 0) -> 
    events_val(Event) := ((Cell//G_S//2) mod 10 =:= 2 -> 3 else 5)
  else
    events_val(Event) :=0
  )),
  fail.
evaluete_events(_, forcast):-
  implication(Event, Ind, Name, P_Event),
  Cell := events2(10*Event),
  Action := ((Cell//G_S//2) mod 10),
  (Cell =:= max_integer ->
    write("Error")
  else (event_is_true(Event, 1, 0), Action<3, is_forcast(Event, Action)->
    (Action>0 ->
      events_val(Event) :=      8*(3-Action)+2*((Cell//2) mod G_S) + Cell mod 2
    else
      events_val(Event) := 32 + 8*(3-1)     +2*((Cell//2) mod G_S) + Cell mod 2
    )
  else
    events_val(Event) :=0
  )),
  fail.
evaluete_events(Views, restriction):-
  implication(Event, Ind, Name, P_Event),
  events_val(Event) :=0,
  Cell := events2(10*Event),
  Action := ((Cell//G_S//2) mod 10),
  (Views=:=1 ->
    Action =:= 2
  else (Views=:=3 ->
    Action < 2
  else
    Action > 2
  )),
  (Cell =:= max_integer ->
    write("Error")
  else (event_is_true(Event, 1, 0)->
    events_val(Event) := 2 + true_value(cell_is_true(Cell, 0))
  )),
  fail.
evaluete_events(_, _).

is_forcast(Event, Action):-
  An := events2(10*Event+1)//G_S//2,
  (Action=:=2 -> An=\=2 else An>1).

events_are_true(Model, Events, Views):-
  Events := 0,
  for(Event, 0, all_arrows(Model)-1),
    Event2 := events(G_T*Model+Event),
    (events_val(Event2) =:= 4-Views ->
      Events := Events \/ 2**Event
    ),
    fail.
events_are_true(Model, Events, Views).

event_is_true(Event, Count, Delay):-
  Cell := events2(10*Event+Count),
  (Cell =\= max_integer ->
    cell_is_true(Cell, Delay),
    event_is_true(Event, Count+1, Delay)
  ).

cell_is_true(Cell, Delay):-
  PosCell := Cell//2,
  Atom := PosCell//G_S//10,
  Type := (PosCell//G_S) mod 10,
  Value := PosCell mod G_S,
  (Type =:= 1 ->
    Res := true_value(buf_incorrect((G_Time+Delay-Atom) mod 50) /\ 2**Value =:= 0)
  else (Type =:= 3 ->
    Res := true_value(active_previous(Atom)/\2**Value=\=0)
  else (Type =:= 4 ->
    Res := true_value(observable(Atom))
  else (Type =:= 6 ->
    Res := events_val(Atom)
  else
    Buffer := (Type =:= 0 -> buf_actions else buf_views),
    Res := true_value(Buffer((G_Time+Delay-Atom) mod 50) =:= Value)
  )))),
  (Res + Cell) mod 2 =:= 1.

